import datetime
import json
import traceback
from typing import Optional

import click
import requests
import urllib3

messages = []
try:
    import config.orc_config as orc_config

    CONFIG_ATTRS = orc_config.__dict__
except ModuleNotFoundError:
    messages.append("Ошибка, отсутствует конфигурационный файл!")

HTTP_401_UNAUTHORIZED = 401
HTTP_404_NOT_FOUND = 404

LEN_MAGIC_METHOD = "__len__"
OPENSTACK_TOKEN = "X-Subject-Token"
TERMIDESK_HOST = "TERMIDESK_HOST"
OPENSTACK = "OpenStack"
TERMIDESK = "Termidesk"
MIN_LEN_PARAM = 1
MIN_LEN_HOST = 3
ORC_SECRET_KEY = "ORC_SECRET_KEY"
MIN_LEN_ORC_SECRET_KEY = 45
BOOL_CORRECT_VALUES = ("True", "False")
CONTENT_TYPE_APPLICATION_JSON = "application/json"
# Адрес с данными доменов аутентификации
TERMIDESK_AUTHENCTICATORS_PATH = "/api/auth/v1/authenticators"
# Путь до файла с crash report
CRASH_REPORT_FILE_PATH = "/tmp/termidesk_self_check_crash.log"
# EX: 2023-01-19 15:45:13.084655
TIME_START = datetime.datetime.now()
EXCEPTION_MESSAGE = f"{TIME_START} Возникла ошибка.\n"
# Атрибут указывает, должен ли файл быть перезаписан.
IS_NEED_CREATE_NEW_FILE = True
# открытие на запись, содержимое файла удаляется, если файла не существует, создается новый.
FILE_MODE_W = "w"
# открытие на дозапись, информация добавляется в конец файла.
FILE_MODE_A = "a"
OPENSTACK_PARAMS = (
    "APP_CREDENTIALS_NAME",
    "APP_CREDENTIALS_SECRET",
    "APP_CREDENTIALS_USER_NAME",
    "APP_CREDENTIALS_USER_DOMAIN",
    "OPENSTACK_AUTH_URL",
)
TERMIDESK_PARAMS = (
    "TERMIDESK_HOST",
    "TERMIDESK_PORT",
    "TERMIDESK_USERNAME",
    "TERMIDESK_PASSWORD",
    "TERMIDESK_AUTH",
    "TERMIDESK_AUTH_URL",
)
DJANGO_PARAMS = (
    "ORC_DEBUG",
    "ORC_SECRET_KEY",
    "ORC_ALLOWED_HOSTS",
    "ORC_LOGLEVEL",
    "ORC_HTTPS_ONLY",
)
DJANGO_FLAGS = (
    "ORC_DEBUG",
    "ORC_HTTPS_ONLY",
)
MESSAGE_NOT_IN_CONFIG_FILE = "Параметр `{param}` отсутствует в конфигурационном файле."
MESSAGE_PARAMETER_EMPTY = "Параметр `{param}` не заполнен."
MESSAGE_VERIFICATION_COMPLETED = f"{TIME_START} Проверка завершена."
MESSAGE_ALL_CHECK_SUCCESSFUL = "Все проверки прошли успешно."
MESSAGE_CRASH_REPORT_GENERAL = (
    f"При работе проверки произошла ошибка. Подробнее: '{CRASH_REPORT_FILE_PATH}'"
)
MESSAGE_CRASH_REPORT = (
    "При попытке подключения к {param} произошла ошибка. "
    + f"Подробнее: '{CRASH_REPORT_FILE_PATH}'"
)
MESSAGE_OPENSTACK_NOT_CONNECTED = (
    f"Не удалось подключиться к OpenStack. {OPENSTACK_TOKEN} отсутствует в заголовках."
)
MESSAGE_INCORRECT_AUTH_URL = (
    "При попытке подключения к {param} произошла ошибка, неверный `{param_url}`."
)
MESSAGE_NOT_CORRECT_HOST_PORT = "При попытке подключения к {param} произошла ошибка, проверьте параметры подключения."
MESSAGE_UNAUTHORIZED = "Произошла ошибка при попытке подключения к {param}, неверные данные аутентификации."
MESSAGE_TERMIDESK_HOST_INCORRECT_LEN = (
    f"Параметр `TERMIDESK_HOST` заполнен неправильно, его длина меньше {MIN_LEN_HOST}."
)
MESSAGE_TERMIDESK_NOT_CONNECTED = (
    "Не удалось подключиться к Termidesk, 'token' отсутствует в теле ответа, "
    "проверьте данные для аутентификации."
)
MESSAGE_TERMIDESK_DOMAIN_CHECK_ERROR = (
    "Не удалось проверить домен аутентификации. "
    + f"Подробнее: '{CRASH_REPORT_FILE_PATH}'"
)
MESSAGE_TERMIDESK_DOMAIN_INCORRECT = (
    "Домен аутентификации отсутствует в списке доменов."
)
MESSAGE_NOT_PARAM_IN_VALUE_LIST = (
    "Параметр `{param}` имеет неправильное значение! "
    + f"Допустимые значения: {BOOL_CORRECT_VALUES}."
)
MESSAGE_DJANGO_SECRET_KEY_HAS_SHORT_VALUE = (
    f"Параметр `{ORC_SECRET_KEY}` имеет слишком короткое значение. "
    + f"Минимальная длина {MIN_LEN_ORC_SECRET_KEY} символов!"
)

# Отключаем предупреждение об игнорировании сертификата
urllib3.disable_warnings()


def is_value_satisfies_min_length(value: str, min_length=MIN_LEN_PARAM) -> bool:
    """
    Проверяет, соответствует ли значение минимальной длине.
    """
    return len(value) >= min_length


def check_param_exists_in_config_file(param):
    """
    Проверяет, что параметр находится в конфиг файле.
    Если параметр отсутствует в конфигурационном файле, записывает информацию в messages.
    """
    if param not in CONFIG_ATTRS:
        messages.append(MESSAGE_NOT_IN_CONFIG_FILE.format(param=param))


def check_param_satisfies_min_length(param):
    """
    Проверяет, что значение параметра соответствует минимальной длине.
    Если значение не соответствует минимальной длине, записывает информацию в messages.
    """
    value = CONFIG_ATTRS.get(param)
    if (
        param in CONFIG_ATTRS
        and hasattr(value, LEN_MAGIC_METHOD)
        and not is_value_satisfies_min_length(value)
    ):
        messages.append(MESSAGE_PARAMETER_EMPTY.format(param=param))


def check_params_are_filled(params):
    """
    Проверяет все параметры на наличие в конфигурационном файле и
    соответствие значения минимальной длине.
    """
    for param in params:
        check_param_exists_in_config_file(param)

        check_param_satisfies_min_length(param)


def exception_message_to_file():
    """
    Записывает traceback в файл CRASH_REPORT_FILE_PATH.
    """
    global IS_NEED_CREATE_NEW_FILE
    with open(
        CRASH_REPORT_FILE_PATH,
        FILE_MODE_W if IS_NEED_CREATE_NEW_FILE else FILE_MODE_A,
    ) as file:
        IS_NEED_CREATE_NEW_FILE = False
        file.write(EXCEPTION_MESSAGE)
        file.write(traceback.format_exc())


def openstack_check_connect():
    """
    Проверяет, что подключение к OpenStack успешно.
    Если соединение не удалось, записывает информацию в messages.
    """
    payload = {
        "auth": {
            "identity": {
                "methods": ["application_credential"],
                "application_credential": {
                    "name": CONFIG_ATTRS["APP_CREDENTIALS_NAME"],
                    "secret": CONFIG_ATTRS["APP_CREDENTIALS_SECRET"],
                    "user": {
                        "name": CONFIG_ATTRS["APP_CREDENTIALS_USER_NAME"],
                        "domain": {"name": CONFIG_ATTRS["APP_CREDENTIALS_USER_DOMAIN"]},
                    },
                },
            }
        }
    }

    headers = {
        "content-type": "application/json",
    }
    try:
        response = requests.post(
            url=CONFIG_ATTRS["OPENSTACK_AUTH_URL"],
            data=json.dumps(payload),
            headers=headers,
            timeout=CONFIG_ATTRS["REQUEST_TIMEOUT"],
            verify=False,
        )
        response.raise_for_status()
        if OPENSTACK_TOKEN not in response.headers:
            messages.append(MESSAGE_OPENSTACK_NOT_CONNECTED)
    except requests.exceptions.ConnectionError:
        messages.append(MESSAGE_NOT_CORRECT_HOST_PORT.format(param=OPENSTACK))
    except requests.exceptions.HTTPError:
        if response.status_code == HTTP_401_UNAUTHORIZED:
            messages.append(MESSAGE_UNAUTHORIZED.format(param=OPENSTACK))
        elif response.status_code == HTTP_404_NOT_FOUND:
            content_type = response.headers["Content-Type"]
            # Если content_type == 'application/json', ожидаем ошибку авторизации, неверное имя,
            # иначе неверный адрес
            if content_type == CONTENT_TYPE_APPLICATION_JSON:
                messages.append(MESSAGE_UNAUTHORIZED.format(param=OPENSTACK))
            else:
                messages.append(
                    MESSAGE_INCORRECT_AUTH_URL.format(
                        param=OPENSTACK,
                        param_url="OPENSTACK_AUTH_URL",
                    )
                )
        else:
            exception_message_to_file()
            messages.append(MESSAGE_CRASH_REPORT.format(param=OPENSTACK))
    except Exception:
        exception_message_to_file()
        messages.append(MESSAGE_CRASH_REPORT.format(param=OPENSTACK))


def check_openstack_config_section(params):
    """
    Функция возвращает словарь возникших несоответствий.
    """
    start_messages_len = len(messages)
    check_params_are_filled(params)
    if start_messages_len == len(
        messages
    ):  # FIXME: временное решение, нужно пересмотреть
        openstack_check_connect()


def termidesk_check_host_correct():
    """
    Проверяет, что параметр "TERMIDESK_HOST" соответствует минимальной длине.
    Если параметр не соответствует, записывает информацию в messages.
    """
    is_correct = TERMIDESK_HOST in CONFIG_ATTRS and is_value_satisfies_min_length(
        CONFIG_ATTRS[TERMIDESK_HOST], MIN_LEN_HOST
    )
    if not is_correct:
        messages.append(MESSAGE_TERMIDESK_HOST_INCORRECT_LEN)


def get_termidesk_auth_domains() -> Optional[list]:
    """
    Получает и возвращает список имен доменов проверки подлинности.
    Ничего не возвращает, если во время запроса возникает ошибка.
    """
    try:
        url = (
            f"http://{CONFIG_ATTRS[TERMIDESK_HOST]}:"
            f"{CONFIG_ATTRS['TERMIDESK_PORT']}{TERMIDESK_AUTHENCTICATORS_PATH}"
        )
        response = requests.get(
            url=url,
            timeout=CONFIG_ATTRS["REQUEST_TIMEOUT"],
        )
        response.raise_for_status()

        return [item["authSmallName"] for item in response.json()]
    except requests.exceptions.ConnectionError:
        messages.append(MESSAGE_NOT_CORRECT_HOST_PORT.format(param=TERMIDESK))
    except Exception:
        exception_message_to_file()
        messages.append(MESSAGE_TERMIDESK_DOMAIN_CHECK_ERROR)


def termidesk_check_auth_domain():
    """
    Проверяет, существует ли введенный домен аутентификации в Termidesk.
    Если параметр отсутствует, записывает информацию в сообщения.
    """
    if "TERMIDESK_AUTH" in CONFIG_ATTRS:
        auth_domains = get_termidesk_auth_domains()
        if auth_domains and CONFIG_ATTRS["TERMIDESK_AUTH"] not in auth_domains:
            messages.append(MESSAGE_TERMIDESK_DOMAIN_INCORRECT)


def termidesk_check_connect():
    """
    Проверяет, что подключение к Termidesk успешно.
    Если соединение не удалось, записывает информацию в messages.
    """
    url = (
        f"http://{CONFIG_ATTRS[TERMIDESK_HOST]}:"
        f"{CONFIG_ATTRS['TERMIDESK_PORT']}{CONFIG_ATTRS['TERMIDESK_AUTH_URL']}"
    )
    payload = {
        "username": CONFIG_ATTRS["TERMIDESK_USERNAME"],
        "password": CONFIG_ATTRS["TERMIDESK_PASSWORD"],
        "authSmallName": CONFIG_ATTRS["TERMIDESK_AUTH"],
    }
    try:
        response = requests.post(
            url=url,
            data=payload,
            timeout=CONFIG_ATTRS["REQUEST_TIMEOUT"],
        )
        response.raise_for_status()
        if "token" not in response.json():
            messages.append(MESSAGE_TERMIDESK_NOT_CONNECTED)
    except requests.exceptions.ConnectionError:
        messages.append(MESSAGE_NOT_CORRECT_HOST_PORT.format(param=TERMIDESK))
    except requests.exceptions.HTTPError:
        if response.status_code == HTTP_401_UNAUTHORIZED:
            messages.append(MESSAGE_UNAUTHORIZED.format(param=TERMIDESK))
        elif response.status_code == HTTP_404_NOT_FOUND:
            messages.append(
                MESSAGE_INCORRECT_AUTH_URL.format(
                    param=TERMIDESK,
                    param_url="TERMIDESK_AUTH_URL",
                )
            )
        else:
            exception_message_to_file()
            messages.append(MESSAGE_CRASH_REPORT.format(param=TERMIDESK))
    except Exception:
        exception_message_to_file()
        messages.append(MESSAGE_CRASH_REPORT.format(param=TERMIDESK))


def check_termidesk_config_section(params):
    """
    Функция возвращает словарь возникших несоответствий.
    """
    start_messages_len = len(messages)
    check_params_are_filled(params)
    termidesk_check_host_correct()
    termidesk_check_auth_domain()
    if start_messages_len == len(
        messages
    ):  # FIXME: временное решение, нужно пересмотреть
        termidesk_check_connect()


def check_bool_param_value_in_value_list(param):
    """
    Проверяет, что значение логического параметра допустимо.
    Если значение не допустимо, записывает информацию в messages.
    """
    param_value = CONFIG_ATTRS.get(param)
    if (
        param in CONFIG_ATTRS
        and isinstance(param_value, str)
        and param_value not in BOOL_CORRECT_VALUES
    ):
        messages.append(MESSAGE_NOT_PARAM_IN_VALUE_LIST.format(param=param))


def check_correct_bool_params(params):
    """
    Проверяет все логические параметры, чтобы убедиться,
    что значение соответствует допустимому списку.
    """
    for param in params:
        check_bool_param_value_in_value_list(param)


def check_django_secret_key():
    """
    Проверяет, что параметр "ORC_SECRET_KEY" соответствует минимальной длине.
    Если параметр не соответствует, записывает информацию в messages.
    """
    param_value = CONFIG_ATTRS.get(ORC_SECRET_KEY)
    if ORC_SECRET_KEY in CONFIG_ATTRS and len(param_value) < MIN_LEN_ORC_SECRET_KEY:
        messages.append(MESSAGE_DJANGO_SECRET_KEY_HAS_SHORT_VALUE)


def check_django_config_section(params, flag_params):
    """
    Функция возвращает словарь возникших несоответствий.
    """
    check_params_are_filled(params)
    check_correct_bool_params(flag_params)
    check_django_secret_key()


@click.command()
def main():
    """
    Проверяет правильность значений всех сред, используемых для работы сервера.

    Валидация проводится для значений: OpenStack, Termidesk, Django.
    """
    try:
        if not messages:
            check_openstack_config_section(OPENSTACK_PARAMS)

            check_termidesk_config_section(TERMIDESK_PARAMS)

            check_django_config_section(DJANGO_PARAMS, DJANGO_FLAGS)
    except Exception:
        exception_message_to_file()
        messages.append(MESSAGE_CRASH_REPORT_GENERAL)
    finally:
        if messages:
            print("\n".join(messages))
        else:
            print(MESSAGE_ALL_CHECK_SUCCESSFUL)
        print(MESSAGE_VERIFICATION_COMPLETED)


if __name__ == "__main__":
    main()
